#include "SDL.h"
#include "SDL_ttf.h"

#include <SDL_pad.h>

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_SFNT_NAMES_H
#include FT_TRUETYPE_IDS_H

#include "fontlist/fontlist.h"

#include <string.h>
#include <assert.h>

/* you can hack SDL_pad inside freely with these APIs */
#include <SDL_padinternal.h>
extern char* Pad_default_bnames[];
extern char* Pad_default_anames[];
extern int Pad_user_button_num;
extern Pad_UserButton* Pad_user_buttons;
extern char* Pad_keynames[-PAD_LAST];

char** bnames;

enum {
    ENGLISH, JAPANESE, LANGUAGES
};
static char* langStr[LANGUAGES] = { "english", "japanese" };

enum {
    MENU_HEAD = 2, MENU_TAIL = 2
};

enum {
    CONFIGURE_BUTTONS = 2
};

enum {
    JOY_NUM = 2
};

static char* trans_table[][LANGUAGES] = {
    { "SDL_pad config tool", "\x53\x44\x4c\x5f\x70\x61\x64\x20\xe8\xa8\xad\xe5\xae\x9a\xe3\x83\x84\xe3\x83\xbc\xe3\x83\xab" },
    { "english", "\xe8\x8b\xb1\xe8\xaa\x9e" },
    { "japanese", "\xe6\x97\xa5\xe6\x9c\xac\xe8\xaa\x9e" },
    { "language", "\xe8\xa8\x80\xe8\xaa\x9e" },
    { "font", "\xe3\x83\x95\xe3\x82\xa9\xe3\x83\xb3\xe3\x83\x88" },
    { "A_BUTTON", "\x41\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "B_BUTTON", "\x42\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "C_BUTTON", "\x43\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "D_BUTTON", "\x44\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "E_BUTTON", "\x45\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "UP_BUTTON", "\xe4\xb8\x8a\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "RIGHT_BUTTON", "\xe5\x8f\xb3\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "DOWN_BUTTON", "\xe4\xb8\x8b\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "LEFT_BUTTON", "\xe5\xb7\xa6\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "END_BUTTON", "\xe7\xb5\x82\xe4\xba\x86\xe3\x83\x9c\xe3\x82\xbf\xe3\x83\xb3" },
    { "Saved!", "\xe4\xbf\x9d\xe5\xad\x98\xe3\x81\x97\xe3\x81\xbe\xe3\x81\x97\xe3\x81\x9f" },
    { "Threshold1", "\xe8\xbb\xb8\xe8\xa8\xad\xe5\xae\x9a\x31" },
    { "Threshold2", "\xe8\xbb\xb8\xe8\xa8\xad\xe5\xae\x9a\x32" },
};

enum {
    WIDTH = 480,
    HEIGHT = 480,
    FONT_SIZE = 14
};

struct font_info {
	char *path;
	char *family;
};

SDL_Surface* scr;
SDL_Surface* surf;
TTF_Font* font;
char* family;

int x, y;
int trans;
int font_index;
int wait;
char msg[256];
int height_max;
int joynum;

SDL_Color fgcol = { 0, 0, 0, 0 };
SDL_Color bgcol = { 255, 255, 255, 0 };
SDL_Color selcol = { 255, 0, 0, 0 };
SDL_Color waitcol = { 0, 0, 255, 0 };

#define FONT_NUM 16
struct font_info font_info_table[FONT_NUM];
int font_info_table_size;

typedef struct {
    Sint16 max;
    Sint16 min;
    Sint16 middle;
} JoystickAdjuster;
JoystickAdjuster joy2adj[JOY_NUM][2];

/* 
 * Unicodeマップと 日本語(MS JAPANESE?) sfnt(何?) フォントを持ってるフォントだと 0 以外
 * そうじゃなかったら 0 
 */

int check_font( FT_Face face )
{
	unsigned int i, charmap_num;
	FT_UInt snamec;

	if ( FT_IS_SFNT(face) == 0 )
		return 0;

	snamec = FT_Get_Sfnt_Name_Count(face);

	for ( i=0; i<snamec; i++ ) {
		FT_SfntName name;
		FT_Get_Sfnt_Name( face, i, &name );
		
		if ( name.language_id == TT_MS_LANGID_JAPANESE_JAPAN )
			break;
		
	}

	if ( i == snamec )
		return 0;

	charmap_num = face->num_charmaps;

	for ( i=0; i<charmap_num; i++ ) {
		FT_CharMap map = face->charmaps[i];

		if ( (map->encoding==FT_ENCODING_UNICODE) )
			return 1;
	}

	return 0;
}

void init_font_info_table()
{
	FontList *fl = fontlist_open();
	FT_Error err;
	FT_Library ftlib;

	int index=0;

	err = FT_Init_FreeType( &ftlib );

	if ( err != FT_Err_Ok ) {
		puts("ftlib error");
		abort();
	}

	while ( index<FONT_NUM ) {
		FT_Face face;
		char *path = fontlist_next_file( fl );

		if ( path == NULL )
			break;

		err = FT_New_Face( ftlib, path, 0, &face );
		if ( err != FT_Err_Ok )
			continue;
		
		if ( check_font(face) == 0)
			goto end;

		font_info_table[index].path = strdup(path);
		font_info_table[index].family = strdup(face->family_name);

		puts( face->family_name );
		puts( path );

		index++;

	end:
		FT_Done_Face( face );

	}

	font_info_table_size = index;

	FT_Done_FreeType( ftlib );
	fontlist_close( fl );
}

void draw_string_col(char* str, int dx, int dy, SDL_Color col) {
    int i;
    SDL_Rect dst;

    if (trans) {
        int trans_num = (int)(sizeof(trans_table)/sizeof(char*[LANGUAGES]));
        for (i = 0; i < trans_num; i++) {
            if (strcmp(str, trans_table[i][0]) == 0) {
                str = trans_table[i][trans];
            }
        }
    }

    surf = TTF_RenderUTF8_Solid(font, str, col);
    dst.x = dx*160+20;
    dst.y = dy*20+60;
    SDL_BlitSurface(surf, 0, scr, &dst);
    SDL_FreeSurface(surf);
}

void draw_string(char* str, int dx, int dy) {
    SDL_Color col = (dx == x && dy == y) ? (wait) ? waitcol : selcol : fgcol;
    draw_string_col(str, dx, dy, col);
}

void draw_string_ln(char* str, int dy) {
    SDL_Color col = (dy == y) ? (wait) ? waitcol : selcol : fgcol;
    draw_string_col(str, 1, dy, col);
}

void draw_screen() {
    char** n;
    int i, j;
    int dy = MENU_HEAD;

    SDL_FillRect(scr, 0, SDL_MapRGB(scr->format, 255,255,255));

    draw_string("SDL_pad config tool", 0, -2);

    draw_string("language", 0, 0);
    draw_string_ln(langStr[trans], 0);

    draw_string("font", 0, 1);
    draw_string_ln(family, 1);

    i = 0;
    n = bnames;
    for (; *n && i < Pad_user_button_num; n++, i++) {
        if (Pad_user_buttons[i][0] == PAD_UNUSED) continue;

        draw_string(*n, 0, dy);

        for (j = 0; j < CONFIGURE_BUTTONS; j++) {
            int b = Pad_user_buttons[i][j];
            char* name = NULL;
            if (b == PAD_UNUSED) break;
            name = Pad_keynames[-b];
            if (name) {
                draw_string(name, j+1, dy);
            }
        }
                
        dy++;
    }
    for (i = 0; i < joynum; i++) {
        char buf[256];
        sprintf(buf, "Threshold%d", i+1);
        draw_string_ln(buf, dy);
        dy++;
    }

    draw_string_ln("Save", dy);
    dy++;

    draw_string_ln("Quit", dy);
    dy++;

    dy++;
    draw_string(msg, 0, dy);

    SDL_UpdateRect(scr, 0,0,0,0);
}

void update_font() {
    if (font) TTF_CloseFont(font);
    font = TTF_OpenFont(font_info_table[font_index].path, FONT_SIZE);
    family = font_info_table[font_index].family;
}

void set_pad(int id) {
    int i;
    int index;
    char** n;

    if (id == PAD_UNUSED) return;

    wait = 0;
    i = 0;
    n = bnames;
    index = y-MENU_HEAD;
    for (; *n && i < Pad_user_button_num; n++, i++) {
        if (Pad_user_buttons[i][0] == PAD_UNUSED) continue;
        if (index == 0) break;
        index--;
    }
    Pad_user_buttons[i][x-1] = id;
}

void set_thre(SDL_Event* ev) {
    int i;
    int id = Pad_PadKeyFromEvent(ev);

    int joy = - (height_max - MENU_TAIL - joynum - y) - 1;
    assert(joy < 2);

    if (ev->type == SDL_JOYAXISMOTION &&
        ev->jaxis.which == joy && ev->jaxis.axis < 2)
    {
        int val = ev->jaxis.value;
        JoystickAdjuster* adj;

        adj = &joy2adj[joy][ev->jaxis.axis];

        if (adj->max < val) adj->max = val;
        if (adj->min > val) adj->min = val;
    }

    for (i = 0; i < 2; i++) {
        JoystickAdjuster* adj = &joy2adj[joy][i];
        int v1, v2;
        v1 = (adj->max * 4 + adj->min) / 5;
        v2 = (adj->max + adj->middle) / 2;
        Pad_thre_plus[joy][i] = (v1 < v2) ? v1 : v2;
        v1 = - (adj->max + adj->min * 4) / 5;
        v2 = - (adj->min + adj->middle) / 2;
        Pad_thre_minus[joy][i] = (v1 < v2) ? v1 : v2;
    }


    sprintf(msg, "LEFT: %d, DOWN: %d, UP: %d, RIGHT: %d",
            Pad_thre_minus[joy][0], Pad_thre_plus[joy][1],
            Pad_thre_minus[joy][1], Pad_thre_plus[joy][0]);

    switch (id) {
    case (PADK_z):
    case (PADK_RETURN):
    case (PADK_SPACE):
    case (PADJ1_BUTTON0):
        wait = 0;
        break;
    }
}

#ifdef main
#undef main
#endif
int main(int argc, char* argv[]) {
    int running = 1;
    int actioned = 1;
    char* padrc;

    int i, j;
    char** n;

    for (i = 0; i < JOY_NUM; i++) {
        for (j = 0; j < 2; j++) {
            joy2adj[i][j].max = 0;
            joy2adj[i][j].min = 0;
            joy2adj[i][j].middle = 0;
        }
    }

    if (argc < 2) {
        padrc = "padrc";
    }
    else {
        padrc = argv[1];
    }

    SDL_Init(SDL_INIT_VIDEO|SDL_INIT_JOYSTICK);
    scr = SDL_SetVideoMode(WIDTH, HEIGHT, 32, SDL_SWSURFACE);

    TTF_Init();

    Pad_Init();

    for (joynum = 0; joynum < JOY_NUM; joynum++) {
        if (SDL_JoystickOpened(joynum) == 0) break;
    }
    printf("%d joysticks detected\n", joynum);

    bnames = Pad_default_bnames;
    if (Pad_Load(padrc, bnames, Pad_default_anames) < 0) {
        printf("%s: cannot load\n", padrc);
        exit(1);
    }

    init_font_info_table();

    x = 1;
    y = 0;
    trans = ENGLISH;
    font_index = 0;
    font = 0;
    update_font();
    wait = 0;
    msg[0] = '\0';

    height_max = 3;
    i = 0;
    n = bnames;
    for (; *n && i < Pad_user_button_num; n++, i++) {
        if (Pad_user_buttons[i][0] == PAD_UNUSED) continue;
        height_max++;
    }
    height_max += joynum;

    while (running) {
        draw_screen();

        {
            SDL_Event ev;
            int id;

            if (actioned) msg[0] = '\0';
            actioned = 1;

            SDL_WaitEvent(&ev);
            id = Pad_PadKeyFromEvent(&ev);

            if (id == PADE_QUIT) {
                running = 0;
                break;
            }

            if (wait) {
                if (height_max - MENU_TAIL - joynum < y) {
                    set_thre(&ev);
                }
                else {
                    set_pad(id);
                }
                continue;
            }

            switch (id) {
            case (PADK_UP):
            case (PADJ1_UP):
                if (y > 0) y--; break;
            case (PADK_DOWN):
            case (PADJ1_DOWN):
                if (y < height_max) y++; break;
            case (PADK_RIGHT):
            case (PADJ1_RIGHT):
                if (y == 0) {
                    if (trans < LANGUAGES-1) trans++;
                }
                else if (y == 1) {
                    if (font_index < font_info_table_size-1) font_index++;
                    update_font();
                }
                else {
                    if (x < CONFIGURE_BUTTONS) x++;
                }
                break;
            case (PADK_LEFT):
            case (PADJ1_LEFT):
                if (y == 0) {
                    if (trans > 0) trans--;
                }
                else if (y == 1) {
                    if (font_index > 0) font_index--;
                    update_font();
                }
                else {
                    if (x > 1) x--;
                }
                break;
            case (PADK_z):
            case (PADK_RETURN):
            case (PADK_SPACE):
            case (PADJ1_BUTTON0):
                if (y == height_max) { /* Quit */
                    running = 0;
                }
                else if (y == height_max - 1) { /* Save */
                    Pad_Save(padrc, bnames, Pad_default_anames);
                    printf("save to %s\n", padrc);
                    sprintf(msg, "Saved!");
                    actioned = 0;
                }
                else if (MENU_HEAD <= y) {
                    wait = 1;
                }
                break;
            case (PADK_s):
                y = height_max-1;
                break;
            case (PADK_ESCAPE):
            case (PADK_q):
                y = height_max;
                break;
            default:
                actioned = 0;
            }
        }
    }

    if (font) TTF_CloseFont(font);

    SDL_Quit();

    return 0;
}
